@@ -706,6 +706,14 @@ struct APIConfig {
 };
 
 struct Config {
+  Config() = default;
+  ~Config();
+
+  Config(Config &&) = delete;
+  Config(const Config &&) = delete;
+  Config &operator=(Config &&) = delete;
+  Config &operator=(const Config &&) = delete;
+
   HttpProxy downstream_http_proxy;
   HttpConfig http;
   Http2Config http2;
@@ -717,13 +725,9 @@ struct Config {
   ImmutableString conf_path;
   ImmutableString user;
   ImmutableString mruby_file;
-  char **original_argv;
-  char **argv;
-  char *cwd;
   size_t num_worker;
   size_t padding;
   size_t rlimit_nofile;
-  int argc;
   uid_t uid;
   gid_t gid;
   pid_t pid;
@@ -736,6 +740,9 @@ struct Config {
 
 const Config *get_config();
 Config *mod_config();
+// Replaces the current config with given |new_config|.  The old config is
+// returned.
+std::unique_ptr<Config> replace_config(std::unique_ptr<Config> new_config);
 void create_config();
 
 // generated by gennghttpxfun.py
@@ -890,10 +897,11 @@ int parse_config(Config *config, const StringRef &opt, const StringRef &optarg,
 int parse_config(Config *config, int optid, const StringRef &opt,
                  const StringRef &optarg, std::set<StringRef> &included_set);
 
-// Loads configurations from |filename| and stores them in statically
-// allocated Config object. This function returns 0 if it succeeds, or
-// -1.  See parse_config() for |include_set|.
-int load_config(const char *filename, std::set<StringRef> &include_set);
+// Loads configurations from |filename| and stores them in |config|.
+// This function returns 0 if it succeeds, or -1.  See parse_config()
+// for |include_set|.
+int load_config(Config *config, const char *filename,
+                std::set<StringRef> &include_set);
 
 // Parses header field in |optarg|.  We expect header field is formed
 // like "NAME: VALUE".  We require that NAME is non empty string.  ":"
